require 'glutil'

createGuiSlider('velDecay', 0, .999, 1)		-- .85 ?
createGuiSlider('densityDecay', 0, .831353, 1)	-- .995 ?
createGuiSlider('timestep', 0, .1, 1)
createGuiSlider('viscousity', 0, .001, 1)
createGuiSlider('cursorDensity', 0, 1.80882, 10)
createGuiSlider('cursorRadius', 0, .158825, 1)
createGuiSlider('cursorVelScale', 0, .01, 10)
createGuiCheckbox('reset', false)
createGuiSlider('diffuseSteps', 1, 3, 20)

--[[
i'm copying so much from flame that i should put all the mutual
code in its own utility lua file
--]]

-- hacked to be lum-based
function getTexture(tex)
	glBindTexture(GL_TEXTURE_2D, tex)
	-- getting an error right here
	local p = glGetTexImage(GL_TEXTURE_2D, 0)	-- i forget what that zero is all about
	local w = glGetTexLevelParameter(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH)
	local h = glGetTexLevelParameter(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT)	
	local ptype = type(p)
	assert(ptype == 'table', 'p got a bad type '..ptype)
	assert(#p == w*h*4, 'i expected #p to be '..(w*h*4)..' but it turns out to be '..#p)
	glBindTexture(GL_TEXTURE_2D, 0)
	return p
end


-- slow copy method
fluidDim = 512
useCopyImage = false
textureModeFloat = false 
textureModeFixed8 = false	-- alignment errors?
textureModeFixed16 = true
doProject = true

function math.clamp(v,min,max)
	return math.min(math.max(v,min), max)
end

--[[
hsv
to recreate this in a shader you just need a 1d texture spanning the hue, with sat and lum set to 100%
to recreate luminance changes, interpolate from the color you're at to black
to recreate saturation changes, interpolate from the color you're at to the luminance greyscale
so in the shader...
vec3 hsv(vec3 hsv) {
	vec3 hue = texture2D(hsvTex, vec2(hsv[0], .5));
	vec3 lum = vec3(hsv[2]);
	return mix(lum, hue, hsv[1]);
	
}

--]]
function hsv2rgb(hsv)
	local r, g, b;
	local i, f, p, q, t;

	local h = (hsv[1] % 1) * 6
	local s = math.clamp(hsv[2], 0, 1)	-- s interpolates the value from grey
	local v = math.clamp(hsv[3], 0, 1)	-- v interpolates the value from black

	i = math.floor(h)	-- which of the 6 hue slices we're in
	f = h - i			-- how much we're in that slice 
	p = v * (1 - s)
	q = v * (1 - s*f)
	t = v * (1 - s*(1-f))

	if i == 0 then
		r = v
		g = t
		b = p
	elseif i == 1 then
		r = q
		g = v
		b = p
	elseif i == 2 then
		r = p
		g = v
		b = t
	elseif i == 3 then
		r = p
		g = q
		b = v
	elseif i == 4 then
		r = t
		g = p
		b = v;
	elseif i == 5 then
		r = v
		g = p
		b = q
	end

	return {r,g,b}
end

-- make me a palette tex as well
hsvTexDetail = 1024
hsvTex = glGenTexture()
glBindTexture(GL_TEXTURE_2D, hsvTex)
glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, hsvTexDetail, 1, 0, GL_RGBA, GL_FLOAT, function(u,v)
	local v = hsv2rgb{u / hsvTexDetail, 1, 1}
	return v[1], v[2], v[3], 1	
end)




vertexCode=[[
varying vec2 tc;
void main() {
	tc = gl_Vertex.xy;
	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
]]

--[[
density:
	addsource
	diffuse
	advect

velocity:
	addsource
	diffuse
	project
	advect
	project
	
combined:
	addsource
	diffuse
	project (velocity only)
	advect
	project (velocity only)
jos stam does density sim then velocity sim.  think it matters that they're separate?
--]]

clearColor={0,0,0,0}
if textureModeFloat then
	clearColor={0,0,0,0}
end
if textureModeFixed8 or textureModeFixed16 then
	clearColor={.5,.5,.5,.5}
end

function createFluidShader(args)
	local prefix
	if textureModeFloat then
		prefix = [[
#define FROM_CELL(v)		v
#define FROM_CELL_X(v)		v
#define FROM_CELL_Y(v)		v
#define TO_CELL(v)			v
#define FROM_DIV(v)			v
#define FROM_PRESSURE(v)	v
#define TO_DIV_PRESSURE(v)	v
]]
	end
	if textureModeFixed8 or textureModeFixed16 then
		prefix = [[
#define MAX_FIXED				1.
#define FROM(v)					(((v) * 2. - 1.) * MAX_FIXED)
#define TO(v)					((v / MAX_FIXED) * .5 + .5)
#define FROM_CELL_X(v)			FROM(v)
#define FROM_CELL_Y(v)			FROM(v)
#define FROM_CELL(v)			FROM(v)
#define TO_CELL(v)				TO(v)
#define FROM_DIV(v)				FROM(v)
#define FROM_PRESSURE(v)		FROM(v)
#define TO_DIV_PRESSURE(v)		TO(v)
]]
	end
	
	return ShaderProgram.new{
		vertexCode=vertexCode;
		fragmentCode=prefix..'\n'..args.code;
		uniforms=args.uniforms;
	}.id
end

-- this shader was made to be run 20x over
-- i'm adding sources in here too.  less passes.
diffuseShader = createFluidShader{
	code=[[
varying vec2 tc;
uniform sampler2D fluidTex;
uniform sampler2D fluidLastTex;
uniform sampler2D sourceTex;
uniform vec2 delta;
uniform float alpha;	//set alpha to dt * fluidDim * fluidDim * viscousity
uniform float dt;		//the infamous
uniform vec4 decay;
void main() {
	//add source
	vec4 source = texture2D(sourceTex, tc);
	source = FROM_CELL(source);

	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	
	//diffuse
	vec4 cell = texture2D(fluidTex, tc);
	cell = FROM_CELL(cell);
	
	//dont' worry about floor()'ing tc to get the correctly aligned texel value post-interpolation
	//	on my cart texture interpolation filters don't work with float textures
	vec4 xp = texture2D(fluidLastTex, tc + dx); xp = FROM_CELL(xp);
	vec4 xn = texture2D(fluidLastTex, tc - dx); xn = FROM_CELL(xn);
	vec4 yp = texture2D(fluidLastTex, tc + dy); yp = FROM_CELL(yp);
	vec4 yn = texture2D(fluidLastTex, tc - dy); yn = FROM_CELL(yn);
	gl_FragColor = (cell + source * dt + alpha * (xp + xn + yp + yn))  / (1. + 4. * alpha);
		
	//maybe decay isn't the right thing to use?
	//after all, (if xp xn yp yn were ref'd from the current buffer and not the last)
	//this is supposed to be a stable solver?  yet it's diverging...
	//...unless those are supposed to be referenced from the buffer two iterations ago...
	gl_FragColor *= decay;
	
	gl_FragColor = TO_CELL(gl_FragColor);
}
]];
	uniforms = {
		fluidTex = 0;
		fluidLastTex = 1;
		sourceTex = 2;
	};
}

advectShader = createFluidShader{
	code=
(function()
	if textureModeFloat then
		return '#define BILINEAR_FILTER_IN_SOFTWARE\n'
	end
	return ''
end)()..
[[

varying vec2 tc;
uniform sampler2D fluidTex;
uniform vec2 delta;
uniform float dt;
void main() {
	vec4 cell = texture2D(fluidTex, tc);
	cell = FROM_CELL(cell);

	//advect
	//floor'd to texel coords
	vec2 cellTc = tc - cell.xy * dt / delta;
#ifdef BILINEAR_FILTER_IN_SOFTWARE		//hardware doesnt work for float texs
	vec2 cellLowerTc = (floor(cellTc / delta) - .5) * delta;
	vec2 cellUpperTc = cellLowerTc + delta;
	//fraction of interpolation along cell in S direction
	vec2 cellFrac = (cellTc - cellLowerTc) / delta;
	//make sure bilinear interpolation is enabled
	//oh wait, it can't be with float textures ..
	//so we either do it in shader code
	//or ... we fixed-point-precision the whole thing ...
	//which would be so fast that i'm considering it
	//especially a 16-bit 
	vec4 v00 = texture2D(fluidTex, vec2(cellLowerTc.x, cellLowerTc.y)); v00 = FROM_CELL(v00);
	vec4 v01 = texture2D(fluidTex, vec2(cellLowerTc.x, cellUpperTc.y)); v01 = FROM_CELL(v01);
	vec4 v10 = texture2D(fluidTex, vec2(cellUpperTc.x, cellLowerTc.y)); v10 = FROM_CELL(v10);
	vec4 v11 = texture2D(fluidTex, vec2(cellUpperTc.x, cellUpperTc.y)); v11 = FROM_CELL(v11);
	vec4 v0 = mix(v00, v01, cellFrac.y);
	vec4 v1 = mix(v10, v11, cellFrac.y);
	gl_FragColor = mix(v0, v1, cellFrac.x);
	gl_FragColor = TO_CELL(gl_FragColor);
#else
	gl_FragColor = texture2D(fluidTex, cellTc);
#endif
}
]];
	uniforms = {
		fluidTex = 0;
	};
}

calcDivShader = createFluidShader{
	code=[[
varying vec2 tc;
uniform sampler2D fluidTex;
uniform vec2 delta;
void main() {
	float h = delta.x;	//1/N in stam's paper, and N is the x and y dimensions
	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	float uxp = texture2D(fluidTex, tc + dx).x; uxp = FROM_CELL_X(uxp);
	float uxn = texture2D(fluidTex, tc - dx).x; uxn = FROM_CELL_X(uxn);
	float vyp = texture2D(fluidTex, tc + dy).y; vyp = FROM_CELL_Y(vyp);
	float vyn = texture2D(fluidTex, tc - dy).y; vyn = FROM_CELL_Y(vyn);
	vec2 uvGrad = vec2(uxp-uxn, vyp-vyn) * delta;
	float div = -.5 * (uvGrad.x + uvGrad.y);
	//div, pressure, 0, 0
	gl_FragColor = vec4(div, 0., 0., 0.);
	gl_FragColor = TO_DIV_PRESSURE(gl_FragColor);
}
]];
	uniforms={
		fluidTex=0;
	};
}

diffusePressureShader = createFluidShader{
	code=[[
varying vec2 tc;
uniform sampler2D divTex;
uniform sampler2D divLastTex;
uniform vec2 delta;
void main() {
	float divLast = texture2D(divTex, tc).x;
	divLast = FROM_DIV(divLast);
	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	float pxp = texture2D(divLastTex, tc + dx).y; pxp = FROM_PRESSURE(pxp);
	float pxn = texture2D(divLastTex, tc - dx).y; pxn = FROM_PRESSURE(pxn);
	float pyp = texture2D(divLastTex, tc + dy).y; pyp = FROM_PRESSURE(pyp);
	float pyn = texture2D(divLastTex, tc - dy).y; pyn = FROM_PRESSURE(pyn);
	float pressure = (divLast + pxp + pxn + pyp + pyn) * .25;
	gl_FragColor = vec4(
		divLast,	//maintain div
		pressure,	//diffuse pressure,
		0., 0.);
	gl_FragColor = TO_DIV_PRESSURE(gl_FragColor);
}
]];
	uniforms={
		divTex=0;
		divLastTex=1;
	}
}

projectShader=createFluidShader{
	code=[[
varying vec2 tc;
uniform sampler2D fluidTex;
uniform sampler2D divTex;
uniform vec2 delta;
void main() {
	vec4 cell = texture2D(fluidTex, tc); cell = FROM_CELL(cell);
	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	float pxp = texture2D(divTex, tc + dx).y; pxp = FROM_PRESSURE(pxp);
	float pxn = texture2D(divTex, tc - dx).y; pxn = FROM_PRESSURE(pxn);
	float pyp = texture2D(divTex, tc + dy).y; pyp = FROM_PRESSURE(pyp);
	float pyn = texture2D(divTex, tc - dy).y; pyn = FROM_PRESSURE(pyn);
	vec2 pGrad = vec2(pxp-pxn, pyp-pyn) / delta;
	gl_FragColor = vec4(cell.xy - .5 * pGrad, cell.zw);
	gl_FragColor = TO_CELL(gl_FragColor);
}
]];
	uniforms={
		fluidTex=0;
		divTex=1;
	};
}

displayShader=createFluidShader{
	code=[[
#define pi ]]..math.pi..'\n'..[[
varying vec2 tc;
uniform sampler2D fluidTex;
uniform sampler2D gradTex;
#define maxSpeed .0001
#define maxDensity .01

vec3 hsv2rgb(vec3 hsv) {
	vec3 hue = texture2D(gradTex, vec2(hsv[0], .5)).xyz;
	vec3 lum = vec3(hsv[2]);
	return mix(lum, hue, hsv[1]);
	
}

void main() {
	vec4 cell = texture2D(fluidTex, tc); cell = FROM_CELL(cell);
#if 0	//velocity in there too ...
	float speed = length(cell.xy);
	cell.xy /= max(speed, maxSpeed);
	cell.xy *= .49;	//[-1,1] -> [0,1]
	cell.xy += .5;
	cell.z /= maxDensity;
	cell.z = fract(cell.z);
	gl_FragColor = vec4(cell.xzy, 1.);
#endif
#if 0	//heat based on density
	gl_FragColor = texture2D(gradTex, vec2(cell.z / maxDensity, .5));
#endif
#if 1
	float speed = sqrt(cell.x * cell.x + cell.y * cell.y);
	float angle = atan(cell.y, cell.x);
	vec3 c = vec3(
		fract((angle / (2. * pi)) + .5),
		1. - fract(speed / maxSpeed),
		fract(cell.z / maxDensity));
	c = hsv2rgb(c);
	gl_FragColor = vec4(c.xyz, 1.);
#endif
}
]];
	uniforms={
		fluidTex=0;
		gradTex=1;
	};
}

addPointSourceShader=createFluidShader{
	code=[[
varying vec2 tc;
uniform sampler2D fluidTex;
uniform vec2 sourcePoint;
uniform vec2 sourceVel;
uniform float sourceDensity;
uniform float sourceRadius;
void main() {
	vec4 cell = texture2D(fluidTex, tc);
	cell = FROM_CELL(cell);
	
	vec2 del = tc - sourcePoint;
	float lenSq = dot(del,del);
	if (lenSq < sourceRadius*sourceRadius) {
		float unitlen = sqrt(lenSq) / sourceRadius;
		float infl = sqrt(1. - unitlen);
		cell.xyz += infl * vec3(sourceVel, sourceDensity);
	}
	
	gl_FragColor = cell;
	gl_FragColor = TO_CELL(gl_FragColor);
}
]];
	uniforms={fluidTex=0;};
}

function drawScreenQuad()
	glBegin(GL_QUADS)
	glTexCoord(0,0)	glVertex(0,0)
	glTexCoord(1,0)	glVertex(1,0)
	glTexCoord(1,1)	glVertex(1,1)
	glTexCoord(0,1)	glVertex(0,1)
	glEnd()
end

fbo = FBO:new()

function drawScreenFBO(args)
	if args.viewport then
		local vp = args.viewport
		glPushAttrib(GL_VIEWPORT_BIT)
		glViewport(vp[1], vp[2], vp[3], vp[4])
	end
	if args.resetProjection then
		glMatrixMode(GL_PROJECTION)
		glPushMatrix()
		glLoadIdentity()
		glOrtho(0,1,0,1,-1,1)
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glLoadIdentity()
	end
	if args.shader then
		glUseProgramObject(args.shader)
	end
	if args.uniforms then
		for k,v in pairs(args.uniforms) do
			glUniformf(k,v)		-- uniformf only, but that still supports vectors =)
		end
	end
	if args.texs then
		for i=1,#args.texs do
			glActiveTexture(GL_TEXTURE0+i-1)
			glBindTexture(GL_TEXTURE_2D, args.texs[i])
		end
	end
	if args.color then
		glColor(args.color)
	end
	fbo:draw2D(args.dst, args.callback or drawScreenQuad)
	if args.texs then
		for i=#args.texs,1,-1 do	-- step -1 so we end up at zero
			glActiveTexture(GL_TEXTURE0+i-1)
			glBindTexture(GL_TEXTURE_2D, 0)
		end
	end
	if args.shader then
		glUseProgramObject(0)
	end
	if args.resetProjection then
		glMatrixMode(GL_PROJECTION)
		glPopMatrix()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
	end
	if args.viewport then
		glPopAttrib()
	end
end

function fboCopy(dst,src,w,h)
	if useCopyImage then
		-- using a simple copyteximage2d copy - fps: 945 (204fps at 20 cycles/frame)
		-- now this sets basetex up as the READ_BUFFER.  usu that is for reading into
		-- in this case its gonna be for reading from into the copyteximage (into the swap buffer)
		-- should the viewport be size,size or not touched or 1,1 or what?
		glPushAttrib(GL_VIEWPORT_BIT)
		glViewport(0,0,w,h)
		glBindTexture(GL_TEXTURE_2D, dst)
		fbo:draw2D(src, function()
			-- now copy texture reads from the current GL_READ_BUFFER
			glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32, 0, 0, w, h, 0)
		end)
		glBindTexture(GL_TEXTURE_2D, 0)
		glPopAttrib()
	else
		-- using a complex render-to-fbo copy - fps: 1064 (555fps at 20 cycles/frame)
		-- set up for fbo
		drawScreenFBO{dst=dst, viewport={0,0,w,h}, resetProjection=true, texs={src}}
	end
end

glEnable(GL_TEXTURE_2D)

function createFloatTex(w,h,gen)
	local id = glGenTexture()
	glBindTexture(GL_TEXTURE_2D, id)
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
	local format
	if textureModeFloat then format = GL_RGBA_FLOAT32 end
	if textureModeFixed8 then format = GL_RGB8 end
	if textureModeFixed16 then format = GL_RGB16 end
	
	glTexImage2D(GL_TEXTURE_2D, 0, format, w, h, 0, GL_RGB, GL_FLOAT, gen)
	return id
end

PingPong = {}
PingPong.__index = PingPong	-- for metatable reference
function PingPong:new(howmany, ...)
	local o = {
		hist = {}
	}
	for i=1,howmany do
		table.insert(o.hist, createFloatTex(...))
	end
	setmetatable(o, PingPong)
	return o
end
function PingPong:swap()
	local tmp = self.hist[#self.hist]
	for i=#self.hist,2,-1 do
		self.hist[i] = self.hist[i-1]
	end
	self.hist[1] = tmp
end
function PingPong:cur() return self.hist[1] end
function PingPong:prev(n) return self.hist[(n or 1)+1] end	-- prev() = prev(1), ...

fluidDimViewport = {0, 0, fluidDim, fluidDim}
fluidDelta = {1/fluidDim, 1/fluidDim}		-- hardcode in shaders?

-- rgba: u,v, density
fluidTexs = PingPong:new(2, fluidDim, fluidDim)

-- rgba: source
sourceTex = createFloatTex(fluidDim, fluidDim, function(u,v)
	return 0,0,0,0
end)

-- rgba: divergence, pressure
divTexs = PingPong:new(2, fluidDim, fluidDim)

function clearTex(dst)
	drawScreenFBO{
		dst=dst;
		color=clearColor;
		resetProjection=true;
		viewport={0,0,fluidDim,fluidDim};
	}
end
function resetAll()
	for i=1,#fluidTexs.hist do
		clearTex(fluidTexs.hist[i])
	end
	for i=1,#divTexs.hist do
		clearTex(divTexs.hist[i])
	end
	clearTex(sourceTex)
end
resetAll()

-- notice - this mucks up args

function diffuse()
	for k=1,diffuseSteps do
		glreport('before diffuse')
		fluidTexs:swap()
		drawScreenFBO{
			dst=fluidTexs:cur();
			shader=diffuseShader;
			texs={fluidTexs:prev(), fluidTexs:prev(), sourceTex};
			uniforms={
				delta = fluidDelta;
				alpha = timestep * fluidDim * fluidDim * viscousity;
				dt = timestep;
				decay = {velDecay, velDecay, densityDecay, 0.};
			};
		}
		glreport('after diffuse')
	end
end

function advect()
	glreport('before advect')
	fluidTexs:swap()
	drawScreenFBO{
		dst=fluidTexs:cur();
		shader=advectShader;
		texs={fluidTexs:prev()};
		uniforms={
			delta=fluidDelta;
			dt=timestep;
		};
	}
	glreport('after advect')
end

function project()
	-- 1) calc div and clear p (in an aux tex somewhere)
	glreport('before calcDiv')
	drawScreenFBO{
		dst=divTexs:cur();
		shader=calcDivShader;
		texs={fluidTexs:cur()};
		uniforms={
			delta=fluidDelta;
		};
	}
	glreport('after calcDiv')
	-- 2) diffuse p ... strangely... perpetually treating div as the source
	for k=1,diffuseSteps do
		glreport('before diffuse pressure')
		divTexs:swap()
		drawScreenFBO{
			dst=divTexs:cur();
			shader=diffusePressureShader;
			texs={divTexs:prev(), divTexs:prev()};
			uniforms={
				delta=fluidDelta;
			};
		}
		glreport('after diffuse pressure')
	end
	-- ... that means converge it until it makes up for the difference of divergences (the pressure!)
	-- 3) 
	glreport('before project')
	fluidTexs:swap()
	drawScreenFBO{
		dst=fluidTexs:cur();
		shader=projectShader;
		texs={fluidTexs:prev(), divTexs:cur()};
		uniforms={
			delta=fluidDelta;
		};
	}
	glreport('after project')
end

function iterate()
	diffuse()
	if doProject then project() end
	advect()
	if doProject then project() end
end

function addMouseSource(pos,vel)
	fluidTexs:swap()
--[[
--shader-less, but i'm suspicious that it can't glcolor points negative...
	drawScreenFBO{
		dst=fluidTexs:cur();
		texs={fluidTexs:prev()};
		callback=function()
			-- screen overlay
			glEnable(GL_TEXTURE_2D)
			glColor(1,1,1,1)
			drawScreenQuad()
			glDisable(GL_TEXTURE_2D)
			-- mouse influence
			glColor(vel[1],vel[2],cursorDensity,1)
			glPointSize(2)
			glEnable(GL_POINT_SMOOTH)
			glBegin(GL_POINTS)
			glVertex(pos)
			glEnd()
			glPointSize(1)
		end;
	}
--]]
	-- with shader, but it works... or does it?
	drawScreenFBO{
		dst=fluidTexs:cur();
		texs={fluidTexs:prev()};
		shader=addPointSourceShader;
		uniforms={
			sourcePoint=pos;
			sourceVel=vel;
			sourceDensity=cursorDensity;
			sourceRadius=cursorRadius;
		}
	}
end

glDisable(GL_DEPTH_TEST)

local lastFieldMousePos = {0,0}
function display()
	glreport('begin display')
	
	if reset then
		resetAll()
		reset = false
	end

	local viewport = glGet(GL_VIEWPORT)
	local aspectRatio = viewport[3] / viewport[4]
	
	-- set up for viewport fbo once
	-- assert that all your fbo targets are the same size
	glViewport(0,0,fluidDim,fluidDim)
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	glOrtho(0,1,0,1,-1,1)
	glMatrixMode(GL_MODELVIEW)
	glLoadIdentity()

	local mouseX,mouseY,mouseButton = mouse()
	local fieldMousePos = { mouseX*aspectRatio - (aspectRatio-1)/2, mouseY }
	local fieldMouseVel = {
		(fieldMousePos[1] - lastFieldMousePos[1]) * cursorVelScale,
		(fieldMousePos[2] - lastFieldMousePos[2]) * cursorVelScale
	}
	if mouseButton ~= 0 then
		addMouseSource(fieldMousePos, fieldMouseVel)
	end
	
	lastFieldMousePos[1] = fieldMousePos[1]
	lastFieldMousePos[2] = fieldMousePos[2]

	iterate()

	-- set up for rendering
	glViewport(unpack(viewport))
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	glOrtho(-(aspectRatio-1)/2, 1+(aspectRatio-1)/2, 0, 1, -1, 1)
	glMatrixMode(GL_MODELVIEW)
	glLoadIdentity()

	glEnable(GL_TEXTURE_2D)
	glUseProgramObject(displayShader)
	glActiveTexture(GL_TEXTURE0)
	glBindTexture(GL_TEXTURE_2D, fluidTexs:cur())
	glActiveTexture(GL_TEXTURE1)
	glBindTexture(GL_TEXTURE_2D, hsvTex)
	--todo - shader to map from velocity / density to some appropriate palette
	drawScreenQuad()
	glUseProgramObject(0)
	glBindTexture(GL_TEXTURE_2D, 0)
	glActiveTexture(GL_TEXTURE0)
	glBindTexture(GL_TEXTURE_2D, 0)
	
	do
		glDisable(GL_TEXTURE_2D)
		glColor(1,0,0)
		glBegin(GL_LINES)
		local cursorSize=.02
		glVertex(fieldMousePos[1]-cursorSize,fieldMousePos[2])
		glVertex(fieldMousePos[1]+cursorSize,fieldMousePos[2])
		glVertex(fieldMousePos[1],fieldMousePos[2]-cursorSize)
		glVertex(fieldMousePos[1],fieldMousePos[2]+cursorSize)
		glEnd()
	end
	
	glreport('end display')
end
